import asyncio
import logging
from typing_extensions import Any

from nioflux.server import __ICON__
from nioflux.pipeline.stage import PipelineStage
from nioflux.pipeline.pipeline import Pipeline
from nioflux.util.readsuntil import readsuntil
from nioflux.util.transport_layer import random_port

DEFAULT_HOST = '0.0.0.0'
DEFAULT_TIMEOUT = 8.0
DEFAULT_BUFFER_SIZE = 65536
DEFAULT_EOT = b'<|eot|>'

logger = logging.getLogger('nioflux.server')


class Server:
    def __init__(self, pipeline: list[PipelineStage],
                 host: str = DEFAULT_HOST, port: int | None = None,
                 timeout: float = DEFAULT_TIMEOUT, buffer_size: int = DEFAULT_BUFFER_SIZE,
                 eot: bytes = DEFAULT_EOT, extra: Any | None = None):
        assert len(pipeline) > 0, 'pipeline is empty'
        self._host = host
        self._port = port if port is not None else random_port()
        self._timeout = timeout
        self._buffer_size = buffer_size
        self._eot = eot
        self._pipeline = pipeline
        self._extra = extra

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def extra(self) -> str:
        return self._extra

    @property
    def pipeline(self) -> list[PipelineStage]:
        return self._pipeline

    @property
    def eot(self) -> bytes:
        return self._eot

    async def _channel_handler(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        _peer_host, _peer_port = writer.get_extra_info('peername')
        logger.debug(f'Channel {_peer_host}:{_peer_port} established.')
        try:
            data = await readsuntil(reader=reader, buffer_size=self._buffer_size,
                                    until=self._eot, timeout=self._timeout)
            _, self._extra, _ = await Pipeline(queue=self._pipeline, data=data, extra=self._extra,
                                               io_ctx=(reader, writer)).launch()
        except TimeoutError:
            logger.warning(f'Read timeout on channel {_peer_host}:{_peer_port}')
        finally:
            writer.close()
            await writer.wait_closed()

    async def run(self):
        server = await asyncio.start_server(client_connected_cb=self._channel_handler,
                                            host=self._host, port=self._port,
                                            limit=self._buffer_size)
        async with server:
            await server.serve_forever()

    def __str__(self) -> str:
        max_len = len(__ICON__.splitlines()[0])
        _host_and_port = f'={self._host}:{self._port}='
        _eot = self._eot.decode('utf-8')
        return f'{__ICON__}{_host_and_port:^{max_len}}'

    def __repr__(self) -> str:
        return (f'nioflux.Server(host={self._host}, port={self._port}, '
                f'timeout={self._timeout}, buffer_size={self._buffer_size}, '
                f'eot={self._eot}, extra={self._extra})')



import asyncio
import logging

from typing_extensions import Any

from nioflux.pipeline.stage import PipelineStage

logger = logging.getLogger('nioflux.pipeline')


class Pipeline:
    def __init__(self, queue: list[PipelineStage], data: Any, extra: Any,
                 io_ctx: tuple[asyncio.StreamReader, asyncio.StreamWriter] | None):
        self._queue = queue.copy()
        self._data = data
        self._extra = extra
        self._err = []
        self._fire = True
        self._io_ctx = io_ctx

    @property
    def io_ctx(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return self._io_ctx

    async def launch(self) -> tuple[Any, Any, list[Exception]]:
        if not isinstance(self._queue, list):
            return self._data, self._extra, self._err
        if len(self._queue) > 0:
            for i, stage in enumerate(self._queue):
                logger.debug(f'Staging ({i + 1}/{len(self._queue)}) {stage.label}')
                self._data, self._extra, self._err, self._fire = await stage(data=self._data, extra=self._extra,
                                                                             err=self._err, fire=self._fire,
                                                                             io_ctx=self._io_ctx)
                if not self._fire:
                    break
        return self._data, self._extra, self._err

    async def __call__(self) -> tuple[Any, Any, list[Exception]]:
        return await self.launch()



import asyncio
import logging
from typing_extensions import Any
from typing_extensions import override

from nioflux.pipeline.stage import PipelineStage

logger = logging.getLogger('nioflux.server')


class ErrorNotify(PipelineStage):
    def __init__(self):
        super().__init__(label='error_notify')

    @override
    async def __call__(self, data: Any, extra: Any, err: list[Exception], fire: bool,
                       io_ctx: tuple[asyncio.StreamReader, asyncio.StreamWriter] | None) -> tuple[Any, Any, list[Exception], bool]:
        _, writer = io_ctx
        _peer_host, _peer_port = writer.get_extra_info('peername')
        if len(err) > 0:
            err_log = f'Error occurred on channel {_peer_host}:{_peer_port}'
            for e in err:
                err_log += f'\n- {type(e).__name__}: {e}'
            logger.error(err_log)
        return data, extra, err, fire



import asyncio
import time
from abc import abstractmethod
from typing_extensions import Any

from vortezwohl.crypt import sha512


class PipelineStage:
    def __init__(self, label: str | None = None):
        if label is None:
            label = sha512(str(time.perf_counter()))[:16]
        self._label: str = label

    @property
    def label(self) -> str:
        return self._label

    @abstractmethod
    async def __call__(self, data: Any, extra: Any, err: list[Exception], fire: bool,
                       io_ctx: tuple[asyncio.StreamReader, asyncio.StreamWriter] | None) -> tuple[Any, Any, list[Exception], bool]:
        # data, extra, err, fire
        ...



import socket
from random import randint


def random_port():
    def port_is_available(_port: int):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(0.1)
        try:
            return s.connect_ex(('127.0.0.1', _port)) != 0
        finally:
            s.close()

    while True:
        _candidate_port = randint(1025, 65535)
        if port_is_available(_candidate_port):
            return _candidate_port


def tcp_send(data: bytes, host: str, port: int, buffer_size: int = 65536, wait: bool = True) -> bytes | None:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.connect((host, port))
            sock.sendall(data)
            if not wait:
                return None
            resp = sock.recv(buffer_size)
            return resp
        except Exception as e:
            raise e



import asyncio
from asyncio import StreamReader


async def readsuntil(reader: StreamReader, buffer_size: int, until: bytes, timeout: float) -> bytes:
    blocks: bytes = bytes(b'')
    while True:
        blocks += await asyncio.wait_for(reader.read(n=buffer_size), timeout)
        if blocks.endswith(until):
            return blocks[:-len(until)]



import asyncio
from typing_extensions import Any
from typing_extensions import override

from nioflux.pipeline.stage import PipelineStage


class StrDecode(PipelineStage):
    def __init__(self):
        super().__init__(label='str_decode')

    @override
    async def __call__(self, data: Any, extra: Any, err: list[Exception], fire: bool,
                       io_ctx: tuple[asyncio.StreamReader, asyncio.StreamWriter] | None) -> tuple[Any, Any, list[Exception], bool]:
        if isinstance(data, bytes):
            return data.decode('utf-8'), extra, err, fire
        else:
            return data, extra, err, fire



__AUTHOR__ = '吴子豪 / Vortez Wohl'
__EMAIL__ = 'vortez.wohl@gmail.com'
__GITHUB__ = 'https://github.com/vortezwohl'
__BLOG__ = 'https://vortezwohl.github.io'

import logging

from .handler import *
from .pipeline import *
from .server import *

logger = logging.getLogger('nioflux')
logger.setLevel(logging.ERROR)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s] %(asctime)s %(name)s : %(message)s')
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logging.getLogger('nioflux.server').setLevel(logging.DEBUG)
logging.getLogger('nioflux.pipeline').setLevel(logging.DEBUG)
